
;*********************************************************************************
;* UDP   transmission and reception    for the AVR family
;*
;* Transmission code from:
;*  	Ing. Igor Cesko
;* 	Slovakia
;* 	cesko@internet.sk
;* 	http://www.cesko.host.sk
;* 	http://www.cesko.euro-science.net
;*
;* with some minor modifications by Pieter-Tjerk de Boer, namely:
;* - made Normal Link Pulses unipolar, as they should be
;* - CRC calculation macro now done inside the EthSendByte macro, rather than separately
;* - some other code reorganization, mostly to make it easy to send not just UDP but also ARP packets
;*
;* Reception code from:
;*  	dr.ir. Pieter-Tjerk de Boer
;*  	pa3fwm@amsat.org
;*  	p.t.deboer@utwente.nl
;*  	http://wwwhome.cs.utwente.nl/~ptdeboer/
;*
;* This code was tested on ATmega328 at 20 MHz, but should work on other ATmegas too, if clocked at that rate
;*
;* Brief description:
;*  Transmission:
;*    IPv4/UDP packets can be sent, each containing 8 bits of user data,
;*    encoded in the 0th, 5th, 10th, ..., 35th byte of the UDP packet's data field.
;*    Effective speed is cca 2.5kByte/s = like 19200 baud serial line.
;*
;*  Reception:
;*    Specially crafted IPv4/UDP packets can be received, containing 20 bytes of user data.
;*    Format of the user data part of such UDP packets: 
;*     8 bytes of 0x55
;*     8 bytes of 0x55 or 0xAA for resp. 0 or 1 as least-significant bit of first user data byte
;*     8 bytes of 0x55 or 0xAA for resp. 0 or 1 as second-least-significant bit of first user data byte
;*     And so on for 6 more user data bits each represented by a groups of 8 bytes in the UDP packet.
;*     This structure of 9*8 bytes repeats 19 times for the other 19 bytes of user data.
;*     Finally another 8 bytes of 0x55.
;*    Note:
;*     Frames may be missed e.g. when they start while the AVR itself is transmitting
;*     or if immediately preceeded by another frame, or simply due to noise.
;*     Garbage may be received; there is no CRC or address checking in the code.
;*     Thus: build some form of sanity checking (e.g. checksum), and retransmission if needed, into your application!
;*
;* Copyright (c) Ing. Igor Cesko 2004
;* Copyright (c) Pieter-Tjerk de Boer, 2014
;* All rights reserved
;*********************************************************************************

.include "../m328def-nopragma.inc"			;microcontroller description file



.equ	EthPort                 =PORTD          ;port for TX Ethernet lines
.equ	EthPortDirection        =DDRD           ;direction port for TX Ethernet lines
.equ	TxPlusPin               =PD0            ;pin for TX- Ethernet line  ; note: the TxPlusPin and TxMinusPin names are the wrong way around! didn't feel like trying to fix this everywhere in the code...
.equ	TxMinusPin              =PD1            ;pin for TX+ Ethernet line
.equ	EthRxPort               =PIND           ;port for RX Ethernet line  ; note: changing this will also require changing some things related to the pin-change-interrupt (see pcint2 etc. in eth_rx.inc)
.equ	EthRxPin                =PD7            ;pin for RX Ethernet line

.equ	LedPort                 =PORTB          ;port for LED
.equ	LedPortDirection        =DDRB           ;direction port for LED
.equ	LedPin                  =PB1            ;pin for LED

.equ	StackBegin		=RAMEND         ;stack


;---------------------------------------------------
; addresses for the packets:

;IP source address (192.168.3.10)
;IP source address (192.168.178.70)
.set IPSourceAddr3=	192
.set IPSourceAddr2=	168
.set IPSourceAddr1=	178
.set IPSourceAddr0=	70

;IP destination address (192.168.3.99 )
;IP destination address (192.168.178.10 )
.set IPDestAddr3=	192
.set IPDestAddr2=	168
.set IPDestAddr1=	178
.set IPDestAddr0=	10

;source UDP port
.equ UDPSourcePort=	6668

;destination UDP port
.equ UDPDestPort=	6668


;------------------------------------------------------------------------------------------
;********************************************************************
;* Interrupt table
;********************************************************************
.cseg
;------------------------------------------------------------------------------------------
;********************************************************************
;* reset
;********************************************************************
.org 0						;po resete
		rjmp	reset
.org PCI2addr
		rjmp	pcint2_handler
.org OVF2addr
		rjmp	OVF2handler

;------------------------------------------------------------------------------------------


.include "eth_rx.inc"
.include "eth_tx.inc"
.include "packets.inc"

;------------------------------------------------------------------------------------------


;------------------------------------------------------------------------------------------
;********************************************************************
;* Timer2 interrupt handler
;********************************************************************
OVF2handler:
		; called every 13 ms
		; for cyclic sending Normal Link Pulse (NLP) to ethernet line
		; and for sending gratuitous ARP replies

		push	r16
		in	r16,SREG		;backup SREG
		push	r16
		sei				;we enable the interrupts as soon as possible again, since reception should have higher priority

		in	r16,GPIOR1		;if positive, GPIOR1 is number of interrupts until next ARP transmission
						;if negative, it's minus the number of interrupts since last ARP transmission
						;stops decrementing at 128 = -128
		subi	r16,1		; note: dec r16 is not equivalent, since we need the carry flag
		brvs	noARPnow	; branch if r16 was 128
		out	GPIOR1,r16
		brcc	noARPnow	; branch if r16 was not 0
                rcall	SendARP

                rjmp	pulsedone
noARPnow:
		rcall	SendNLP			;otherwise, send Normal Link Pulse
pulsedone:

;		in r16,GPIOR1		; uncomment these lines to send gratuitous ARP even if there's no traffic, useful for testing
;		andi r16,127		; 
;		andi r16,7
;		out GPIOR1,r16

		pop	r16
		out	SREG,r16		;restore SREG
		pop	r16
		reti


;------------------------------------------------------------------------------------------
;********************************************************************
;* Init program
;********************************************************************
;------------------------------------------------------------------------------------------
reset:			;initialization of processor and variables
		ldi	temp0,LOW(StackBegin)	;stack initialization
		out	SPL,temp0
                ldi	temp0,HIGH(StackBegin)	;
		out	SPH,temp0

		ldi	temp0,(1<<PCIE2)	; enable pin change interrupt for the receive line
		sts	PCICR,temp0
		ldi	temp0,(1<<EthRxPin)
		sts	PCMSK2,temp0

		ldi	temp0,7			;set clock input to CLK/1024 = clock every 0.0512ms for Timer2
		sts	TCCR2B,temp0
		ldi	temp0,1<<TOIE2		;enable interrupt from timer2 (every 256 tick = cca13ms= 256*0.0512ms)
		sts	TIMSK2,temp0


		ldi	XH,1		; X pointer points in SRAM, where we'll store intermediate received data
		ldi	XL,0		; we'll use less than 256 bytes, so only change XL
		ldi	YH,2		; Y pointer points in SRAM, where we'll store received user data bytes
		ldi	YL,0		; we'll use less than 256 bytes, so only change YL

	        ldi r16,(1<<LedPin)
                out LedPort,r16             ; enable LED
                out LedPortDirection,r16    ; enable LED

                sei

;------------------------------------------------------------------------------------------
;********************************************************************
;* Main program
;********************************************************************
; The Main program doesn't have much to do, for a minimal application:
; - periodically (once every 8 to 24 ms) sending a Normal Link Pulse is done by the timer 2 handler.
; - sending a (possibly gratuitous) ARP packet is also done by timer 2, in response to receiving a too short packet (which may be an ARP request).
; - responding to incoming packets is done in the received packet handler.
; So nothing left for the main loop in this simple case :-)
 
Main:						;Main program loop
		rjmp Main


;------------------------------------------------------------------------------------------
;********************************************************************
;* received packet handler
; - called when packet has been received, from within interrupt handler, and with interrupts still disabled!
; - data is in memory pointed to by Y, for YL=0..19
; - the example code toggles a led, and sends back all 20 received data bytes, one by one
;********************************************************************

rx_success:
		call toggleled

                inc r30
                mov temp2, r30
		rcall	SendOneUDPByte		; send r30

		ldi yl,0
L02:		ld temp2,y+
		rcall	SendOneUDPByte		; send back all 20 received bytes via UDP
		cpi yl,20
		brne L02

		ret


toggleled:
		in r16,LedPort
	        ldi r17,(1<<LedPin)
	        eor r16,r17
                out LedPort,r16   ; toggle led
		ret


;------------------------------------------------------------------------------------------
;********************************************************************
;* Main program END
;* 
;* What remains it the subroutines transmitting the actual packets.
;* they go at the end since they generate very long code, which would
;* otherwise hamper the use of rjmp/rcall in the rest of the code.
;********************************************************************


;MAC dest (broadcast = FF-FF-FF-FF-FF-FF)
.set MACDestAddr5=	0xFF
.set MACDestAddr4=	0xFF
.set MACDestAddr3=	0xFF
.set MACDestAddr2=	0xFF
.set MACDestAddr1=	0xFF
.set MACDestAddr0=	0xFF

;MAC source (02-01-23-45-67-89)		
; note: the last bit of the first byte should be 0 to mark it as unicast
; note: the 2nd-last bit of the first byte should be 1 to mark it as "locally administered"
.equ MACSourceAddr5=	0x02
.equ MACSourceAddr4=	0x01
.equ MACSourceAddr3=	0x23
.equ MACSourceAddr2=	0x45
.equ MACSourceAddr1=	0x67
.equ MACSourceAddr0=	0x89



;------------------------------------------------------------------------------------------
SendARP:
                send_packet_arp			; call macro to do the actual work
		ret


;------------------------------------------------------------------------------------------
SendOneUDPByte:
; sends byte in register temp2
                send_packet_udp			; call macro to do the actual work
		ret
		
;------------------------------------------------------------------------------------------

; can add more destinations to send UDP packets to, by .set'ing other IP and MAC addresses and then invoking send_packet_udp again

.set IPDestAddr3=	1
.set IPDestAddr2=	2
.set IPDestAddr1=	3
.set IPDestAddr0=	4

.set MACDestAddr5=	0x88
.set MACDestAddr4=	0x9f
.set MACDestAddr3=	0xfa
.set MACDestAddr2=	0xd6
.set MACDestAddr1=	0x3b
.set MACDestAddr0=	0x3b

SendOneUDPByte_second_destination:
; sends byte in register temp2 to alternate destination
                send_packet_udp			; call macro to do the actual work
		ret
		


;------------------------------------------------------------------------------------------
;********************************************************************
;* End of file
;********************************************************************
